# pylint: disable=W0201, W0221, W0223
"""
    A json rpc 2.0
"""
import uuid
from tornado.web import HTTPError
from tornado.websocket import WebSocketHandler as Ws
from .json_utils import dumps, loads
from .json_rpc_mixin import JsonRpcMixin as RpcMix
from .broadcast_mixin import BroadcastMixin as BcMix
from .user_mixin import UserMixin as UMix


class RpcWebsocket(UMix, BcMix, RpcMix, Ws):
    """ Websocket Rpc Class """

    @property
    def procedures(self):
        """ what we make available """
        return self.settings["procedures"]

    def open(self, *args, **kwargs):
        """ websocket open - register user """
        self._id = str(uuid.uuid4())
        if self.current_user is None and self.settings.get("login_url"):
            raise HTTPError(403)
        self.init_broadcast()

    async def on_message(self, data):
        """ handle the action """
        content = loads(data)
        result = await self.handle_content(content)
        if result:
            self.write_message(dumps(result))

    def on_close(self):
        """ unregister user """
        self.end_broadcast()



# pylint: disable=W0201
"""
    Broadcast function  without Websockets
"""
import asyncio
import logging
import time
import inspect
import tornado.iostream
from tornado import ioloop
from tornado.queues import Queue
from tornado import gen
from .broadcaster import Broadcaster
from . import json_utils

LOGGER = logging.getLogger(__name__)


class RedisBroadcaster(Broadcaster):
    """ simple re-broadcaster """

    def broadcast(self, document):
        """ sends to _clients_ """
        data, user_ids = json_utils.loads(document)
        BroadcastMixin.send(data, user_ids)

    async def send(self, data, user_ids):
        """ sends to redis """
        document = json_utils.dumps([data, user_ids])
        await self.publish(document)


class BroadcastMixin:
    """
        Queue of messages to send to client
    """

    _clients_ = []
    _broadcaster_ = None
    _cron_ = None
    _loop_ = None

    def init_broadcast(self):
        """
            Set up response headers and prepare
            local queue and add self to clients
        """
        self.broadcasting = True
        self.queue = Queue()
        self._clients_.append(self)
        self._task_ = self._loop_.call_later(0, self.do_broadcast)

    async def do_broadcast(self):
        """ this waits on the queue and write_message each data """
        try:
            while self.broadcasting:
                try:
                    delta = time.time() + 0.1
                    data = await self.queue.get(timeout=delta)
                    if inspect.iscoroutinefunction(self.write_message):
                        await self.write_message(data)
                    else:
                        self.write_message(data)
                except gen.TimeoutError:
                    continue
        except asyncio.CancelledError:
            pass
        except tornado.iostream.StreamClosedError:
            pass
        self._task_ = None
        self.end_broadcast()

    def end_broadcast(self):
        """ remove self from clients """
        self.broadcasting = False
        if self in self._clients_:
            self._clients_.remove(self)

    @classmethod
    def broadcast(cls, data, user_ids=None):
        """ thread safe """
        if cls._loop_:
            if cls._broadcaster_:
                cls._loop_.call_later(0, cls._broadcaster_.send, data, user_ids)
            else:
                cls._loop_.call_later(0, cls.send, data, user_ids)

    @classmethod
    def keep_alive(cls):
        """ pings all connected clients """
        logging.debug("keep alive")
        msg = str(time.time())
        for client in cls._clients_:
            client.ping(msg)

    @classmethod
    def send(cls, data, user_ids):
        """ does the actual sending to all _clients_ """
        message = json_utils.dumps(data)
        LOGGER.debug("sending: %s", message)
        for client in cls._clients_:
            if user_ids is None or client.current_user["id"] in user_ids:
                client.queue.put_nowait(message)

    @classmethod
    def init_broadcasts(cls, io_loop, topic_name, redis_url):
        """ set up keep alive and re-broadcaster """
        cls._loop_ = io_loop
        cls._cron_ = ioloop.PeriodicCallback(cls.keep_alive, 30000)
        cls._cron_.start()

        if redis_url:
            cls._broadcaster_ = RedisBroadcaster(topic_name, redis_url)
            cls._loop_.call_later(0, cls._broadcaster_.subscribe)
            LOGGER.info("redis broadcast")
        else:
            cls._broadcaster_ = None
            LOGGER.info("local broadcast")

    @classmethod
    async def tidy_up(cls):
        """ clean up keep alive, broadcaster and client tasks """
        if cls._cron_:
            cls._cron_.stop()
        if cls._broadcaster_:
            await cls._broadcaster_.unsubscribe()
        tasks = [b._task_ for b in cls._clients_ if b._task_]  # pylint: disable=W0212
        for task in tasks:
            task.cancel()
        await asyncio.sleep(0.11)



"""
    Proxy to procedures

    ...because it walks like a duck....

    should you wish to run locally::

        async with Moo(Config) as cow:
            result = await cow.add(2, 2)
            assert result == 4

"""
import logging
from importlib import import_module
from tornado.ioloop import IOLoop
from . import context
from .config import Config

LOGGER = logging.getLogger(__name__)


class Moo:
    """ we want a proxy """

    def __init__(self, config: Config, io_loop: IOLoop = None, user=None):
        self._rpc_ = import_module(config.procedures)
        self._loop_ = io_loop if io_loop else IOLoop.current()
        context.LOOP = self._loop_
        self._user_ = user

    async def __aenter__(self):
        """ act as an async context manager """
        return self

    async def __aexit__(self, type_, value, traceback):
        """ act as an async context manager """

    def __getattribute__(self, name):
        """ public attributes are proxies to procedures """
        LOGGER.debug("attribute %s", name)
        if name[0] == "_":
            return super().__getattribute__(name)
        rpc = self._rpc_
        if name in rpc.__all__:
            proc = getattr(rpc, name)
            assert proc, f"no such procedure {name}"

            def _wrapped_(*args, **kwargs):
                LOGGER.info("calling %s(*%r, **%r)", name, args, kwargs)
                return context.perform(self._user_, proc, *args, **kwargs)

            return _wrapped_
        raise AttributeError(f"no such procedure: {name}")



""" task to create Dockerfile and docker-compose.yml """
import os
import logging
from invoke import task

LOGGER = logging.getLogger(__name__)


@task
def docker(_, name, force=False):
    """ create a config folder with dev, staging & prod yamls """
    docker_file = "Dockerfile"
    LOGGER.info("gen docker: Dockerfile")
    if os.path.isfile(docker_file) and force is False:
        print("that file already exists: ", docker_file)
    else:
        with open(docker_file, "w") as file:
            file.write(
                """# Use an official Python runtime as a parent image
FROM python:3-slim
RUN apt-get update && apt-get install -y git

# Set the working directory to /app
WORKDIR /app

# Copy the current directory contents into the container at /app
COPY . /app

# upgrade pip to make sure it can use git repos
RUN pip install --upgrade pip
RUN pip install liteblue

# Install any needed packages
RUN [ -f requirements.txt ] && pip install -r requirements.txt ||  echo "No requirements.txt"

# Make port 80 available to the world outside this container
EXPOSE 80

# Define environment variable
ENV PYTHONPATH ".:${PYTHONPATH}"
ENV PORT 80

"""
            )
        LOGGER.info("gen docker:docker-compose.yml")
        with open("docker-compose.yml", "w") as file:
            file.write(
                f"""version: '3'
services:
  web:
    build: .
    volumes:
        - .:/app
    image: {name}:latest
    container_name: {name}-web
    command: liteblue run {name}
    ports:
      - "8080:80"
"""
            )



""" context for procedure calls """
import contextvars
import contextlib
import inspect
import logging
from typing import List
from functools import partial
from .broadcast_mixin import BroadcastMixin

LOGGER = logging.getLogger(__name__)

USER = contextvars.ContextVar("USER")

LOOP = None


def current_user() -> dict:
    """ returns the user(actor) for this procedure call """
    return USER.get(None)


def broadcast(data: dict, user_ids: List[int] = None):
    """ broadcast to WebSockets and/or EventSources """
    BroadcastMixin.broadcast(data, user_ids)


@contextlib.contextmanager
def _user_call_(user=None):
    """ With this we setup contextvars and reset """
    utoken = USER.set(user)
    try:
        yield
    finally:
        USER.reset(utoken)


class Performer:  # pylint: disable=R0903
    """ tidy namespace """

    @classmethod
    def _perform_(cls, user, func):
        """ perform a function with user in context """
        with _user_call_(user):
            return func()

    @classmethod
    async def _aperform_(cls, user, func):
        """ perform an async function with user in context """
        with _user_call_(user):
            return await func()

    @classmethod
    def perform(cls, user, proc, *args, **kwargs):
        """ perform in LOOP.executor or await """
        todo = partial(proc, *args, **kwargs)
        LOGGER.info("calling: %s", todo)

        if inspect.iscoroutinefunction(proc):
            result = cls._aperform_(user, todo)
        else:
            result = LOOP.run_in_executor(None, cls._perform_, user, todo)
        return result


def perform(_user_, proc, *args, **kwargs):
    """
        perform a sync or async function and returns a task.
        Sync function are performed in a ThreadPoolExecutor.
        Async functions are performed in this loop.
    """
    return Performer.perform(_user_, proc, *args, **kwargs)



"""
    I thought we might use configparser or yaml
    but then thought better of it. ConfigParser is
    too constricted in the values you can get back.
    Yaml just creates dictionaries and does do
    much else. So sticking with classes for present.
"""
import configparser
from importlib import import_module
from pkg_resources import resource_filename
from invoke import task


@task(help={"project": "name of project"})
def cfg(_, project):
    """ print out the values of the config class of a project """
    cfg_ = import_module(project + ".config")
    for key in dir(cfg_.Config):
        if key[0] != "_":
            print(f"{key}:", getattr(cfg_.Config, key))


@task(help={"project": "name of project"})
def read(_, project):
    """ print config """
    config = configparser.ConfigParser(
        interpolation=configparser.ExtendedInterpolation()
    )
    config.read(f"{project}.ini")
    for section in config.sections():
        print(section)
        for key in config[section]:
            print("\t", key, config[section][key])
    print(dict(**config["tornado"]))


@task(help={"project": "name of project"})
def write(_, project):
    """ write config ini file for project """
    default_ini = resource_filename("liteblue.apps", "default.ini")
    config = configparser.ConfigParser(
        interpolation=configparser.ExtendedInterpolation()
    )
    config.read(default_ini)
    config.set("default", "name", project)
    config.set("default", "static_path", resource_filename("liteblue.apps", "static"))
    with open(f"{project}.ini", "w") as configfile:
        config.write(configfile)



# pylint: disable=W0201, W0223
"""
    A json rpc 2.0
"""
import logging
from tornado.web import RequestHandler, HTTPError
from .json_rpc_mixin import JsonRpcMixin as RpcMix
from .user_mixin import UserMixin as UMix
from . import json_utils

LOGGER = logging.getLogger(__name__)


class RpcHandler(UMix, RpcMix, RequestHandler):
    """ marshall json rpc 2.0 requests and respond """

    @property
    def procedures(self):
        """ what we make available """
        return self.settings["procedures"]

    async def post(self):
        """ We only handle the post method """
        if self.settings.get("login_url") and self.current_user is None:
            raise HTTPError(403)
        content = json_utils.loads(self.request.body)
        result = await self.handle_content(content)
        LOGGER.debug("result: %s", result)
        if result:
            self.set_header("Content-Type", "application/json; charset=UTF-8")
            self.write(json_utils.dumps(result))



"""
    We want a description of functions and classes made remotely callable
"""
import inspect
import typing
from collections import OrderedDict


def annotation_to_str(annotation):
    """ converts annotation to string """
    if annotation is None:
        return None
    if isinstance(annotation, str):
        return annotation
    if inspect.isclass(annotation):
        return annotation.__name__
    if isinstance(annotation, typing._GenericAlias):  # pylint: disable=W0212
        return str(annotation)
    if inspect.isfunction(annotation):
        return annotation.__name__
    return annotation


def describe_sql(metadata):
    """
        returns a jsonable sql metadata
    """
    result = {}
    for table_name in metadata.tables:
        table = metadata.tables[table_name]
        pks = [col.name for col in table.primary_key]
        fkeys = {
            col.column_keys[0]: str(col.elements[0].column)
            for col in table.foreign_key_constraints
        }
        result[table.name] = {
            col.name: {
                "pk": col.name in pks,
                "type": col.type.python_type.__name__,
                "sql": repr(col.type),
                "fkey": fkeys.get(col.name),
            }
            for col in table.c
        }
    return result


def describe(target, limit):
    """
        Will dir target and inspect each function

        functions starting with '_' are ignored,
        as are functions whose name is not in limit
    """
    result = OrderedDict()
    for key in filter(lambda x: x[0] != "_", dir(target)):
        if limit and key not in limit:
            continue
        func = getattr(target, key)
        if inspect.isfunction(func):
            name = key
            desc = inspect.signature(func)
            docs = inspect.getdoc(func)
            result[name] = OrderedDict(
                [
                    ("params", [str(p) for p in desc.parameters.values()]),
                    (
                        "returns",
                        annotation_to_str(desc.return_annotation)
                        if desc.return_annotation is not desc.empty
                        else "",
                    ),
                    ("docs", docs),
                ]
            )
    return result



"""
    Common functions used by Click and Invoke
"""
import os
import logging
import logging.config
from importlib import import_module
from contextlib import contextmanager

LOGGER = logging.getLogger(__name__)


class Colored:
    """ sometimes you need to colour console output """

    @staticmethod
    def green(text):
        """ turns text green """
        return f"\033[92m{text}\033[0m"

    @staticmethod
    def red(text):
        """ turns text red """
        return f"\033[91m{text}\033[0m"

    @staticmethod
    def blue(text):
        """ turns text blue """
        return f"\033[94m{text}\033[0m"

    @staticmethod
    def cyan(text):
        """ turns text cyan """
        return f"\033[36m{text}\033[0m"

    @staticmethod
    def gray(text):
        """ turns text gray """
        return f"\033[90m{text}\033[0m"

    @staticmethod
    def pink(text):
        """ turns text pink """
        return f"\033[33m{text}\033[0m"


@contextmanager
def pid_file(path):
    """ creates a file with the process id and removes it """
    pid = os.getpid()
    with open(path, "w") as file:
        file.write(f"{pid}")
    try:
        yield
    finally:
        if os.path.isfile(path):
            os.unlink(path)


def confirm_action(action):
    """ Are you sure? """
    choice = ""
    while choice not in ["y", "n"]:
        choice = input(f"{action} [Y/N] ").lower()
    return choice == "y"


def qname_to_class(qname):
    """ returns a class from module_name:class_name """
    if isinstance(qname, str):
        _modules, _class = qname.split(":")
        return getattr(import_module(_modules), _class)
    return qname



""" database tasks """
from importlib import import_module
from invoke import task
import alembic.config
from alembic import command
from .utils import confirm_action, Colored


def alembic_cfg(name, settings):
    """ get the alembic config from web config data """
    db_url_name = f"{name}_db_url" if name != "default" else "db_url"
    script_location = (
        f"{name}_alembic_script_location"
        if name != "default"
        else "alembic_script_location"
    )
    cfg = alembic.config.Config()
    cfg.set_main_option("script_location", getattr(settings, script_location))
    cfg.set_main_option("sqlalchemy.url", getattr(settings, db_url_name))
    return cfg


def confirm_db_action(action, db_url):
    """db are you sure? """
    return confirm_action(f"{action} {db_url}")


@task
def upgrade(_, project, revision="head", name="default", force=False):
    """ upgrades db to head or revision """
    if force is True or confirm_db_action(Colored.green("UPGRADE"), name):
        app = import_module(project)
        command.upgrade(alembic_cfg(name, app.config.Config), revision)


@task
def downgrade(_, project, revision="base", name="default"):
    """ downgrades db to base or revision """
    if confirm_db_action(Colored.red("DOWNGRADE"), name):
        app = import_module(project)
        command.downgrade(alembic_cfg(name, app.config.Config), revision)


@task
def revise(_, project, comment, name="default"):
    """ creates an alembic database revision """
    app = import_module(project)
    command.revision(alembic_cfg(name, app.config.Config), comment, autogenerate=True)



""" Redis publish and subscribe in one class """
import logging
import aioredis

LOGGER = logging.getLogger(__name__)


class Broadcaster:
    """ both publish and subscribe to same topic """

    def __init__(self, topic_name: str, redis_url: str):
        self.topic_name = topic_name
        self.redis_url = redis_url
        self.subscription = None

    def broadcast(self, document):
        """ sub class responsibility """

    async def publish(self, document):
        """ publishes a document to topic """
        redis = await aioredis.create_redis_pool(self.redis_url)
        await redis.execute("publish", self.topic_name, document)
        LOGGER.debug("published %s -> %r", self.topic_name, document)
        redis.close()
        await redis.wait_closed()

    async def subscribe(self):
        """ run forever coroutine that listens to topic and broadcasts """
        self.subscription = await aioredis.create_redis(self.redis_url)
        response = await self.subscription.subscribe(self.topic_name)
        channel = response[0]
        LOGGER.info("subscribed to: %s", self.topic_name)
        try:
            while await channel.wait_message():
                document = await channel.get(encoding="utf-8")
                LOGGER.debug("got document: %s", document)
                self.broadcast(document)
        finally:
            self.subscription.close()

    async def unsubscribe(self):
        """ close subscription client """
        if self.subscription:
            self.subscription.close()
            await self.subscription.wait_closed()



"""
    Support for Tornado Authentication.
"""
import hashlib
from sqlalchemy.sql import select, insert, update, and_
from .connection import ConnectionMgr
from .user import user


def login(email: str, password: str):
    """ hash the password and see if you can find the user """
    hex_key = hashlib.md5(password.encode("utf-8")).hexdigest()
    with ConnectionMgr.session() as session:
        row = session.execute(
            select([user], and_(user.c.email == email, user.c.password == hex_key))
        ).first()
        result = None
        if row:
            result = dict(row)
            del result["password"]
        return result


def register(email: str, password: str):
    """ hash the password and save the new user """
    if len(password) < 5:
        raise Exception("Password must be five or more characters")
    hex_key = hashlib.md5(password.encode("utf-8")).hexdigest()
    with ConnectionMgr.session() as session:
        where = user.c.email == email
        row = session.execute(select([user], where)).first()
        if row is None:
            user_dict = {"email": email}
        else:
            if row["password"]:
                raise Exception("Already registered")
            user_dict = dict(row)
        user_dict["password"] = hex_key
        if user_dict.get("id"):
            where = user.c.id == user_dict["id"]
            values = {k: v for k, v in user_dict.items() if k != "id"}
            session.execute(update(user, where, values))
        else:
            user_dict["id"] = session.execute(
                insert(user, user_dict)
            ).inserted_primary_key[0]
        session.commit()
        del user_dict["password"]
        return user_dict



""" A connection to a sqlalchemy db """
import logging
import threading
from contextlib import contextmanager
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

LOGGER = logging.getLogger(__name__)


class NoConnection(Exception):
    """ raised on missing execption """


@contextmanager
def scoped_session(session_cls):
    """Provide a transactional scope around a series of operations."""
    session = session_cls()
    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()


class ConnectionMgr:
    """

    Provides thread safe named access to sqlalchemy sessions

    first declare your connection::

        ConnectionMgr.connection("default", "sqlite:///simple.db")

    then connect::

        with ConnectionMgr.session("default") as session:
            session.execute(sql.insert(role, {"name": "user"}))
            session.commit()

    """

    connections = {}
    connections_lock = threading.Lock()

    @classmethod
    def connection(cls, name, db_url, **kwargs):
        """
        creates a cached engine and session_cls for db_url
        kwargs are passed on the create_engine
        """
        with cls.connections_lock:
            conn = cls.connections.get(name)
            if conn is None:
                engine = create_engine(db_url, **kwargs)
                session_cls = sessionmaker(bind=engine)
                conn = engine, session_cls
                cls.connections[name] = conn
                LOGGER.info("%s connection available", name)
        return conn[0], conn[1]

    @classmethod
    def session(cls, name="default"):
        """
        returns a contextmanager awair session
        with autocommit disabled.
        """
        conn = cls.connections.get(name)

        if conn is None:
            raise NoConnection(f"No connection {name} - check configuration")

        return scoped_session(conn[1])



# pylint: disable=W0201, W0223
"""
    Broadcast function  without Websockets
"""
import logging
from tornado.web import RequestHandler, authenticated
from .broadcast_mixin import BroadcastMixin
from .user_mixin import UserMixin as UMix
from . import json_utils

LOGGER = logging.getLogger(__name__)


class EventSource(UMix, BroadcastMixin, RequestHandler):
    """
        Queue of messages to send to client
    """

    def initialize(self):
        """
            Set up response headers and prepare
            local queue and add self to clients
        """
        self.set_header("content-type", "text/event-stream")
        self.set_header("cache-control", "no-cache")
        self.init_broadcast()
        LOGGER.debug("init")

    @authenticated
    async def get(self):
        """ handle a connection """
        LOGGER.debug("get")
        await self.flush()
        self.init_broadcast()

    async def write_message(self, data):
        """ to look like a websocket for BroadcastMixin """
        self.write(f"data: {data}\n\n")
        await self.flush()
        LOGGER.debug("wrote %s", data)

    def ping(self, msg):
        """ handle keep alive """
        self.queue.put_nowait(json_utils.dumps({"action": "ping", "message": msg}))



# pylint: disable=R0903
""" application configuration """
import os
from pkg_resources import resource_filename


class Config:
    """ Access to configuration """

    name = "liteblue-app"
    port = int(os.getenv("PORT", "8080"))
    max_workers = 5
    tornado_debug = False
    tornado_cookie_name = f"{name}-user"
    tornado_cookie_secret = f"it was dark and stormy night for {name}"
    tornado_login_url = "/login"
    login_page = "login.html"
    static_path = resource_filename("liteblue.apps", "static")
    procedures = ".procedures"

    SQLITE_CONNECT_ARGS = {"echo": False, "encoding": "utf-8"}

    MYSQL_CONNECT_ARGS = {
        "echo": False,
        "isolation_level": "READ COMMITTED",
        "pool_recycle": 120,
        "pool_size": max_workers,
        "encoding": "utf-8",
    }
    alembic_script_location = "alembic"
    db_url = os.getenv("DB_URL", f"sqlite:///{name}.db")
    connection_kwargs = SQLITE_CONNECT_ARGS

    redis_url = None  # if mutiple hosts or worker this is how they communicate
    redis_topic = f"{name}-broadcast"  # topic for all broadcasts
    redis_queue = f"{name}-work"  # queue for remote workers
    redis_workers = 0  # if not 0 then redis workers waiting on redis_queue



# pylint: disable=W0201
"""
    A json rpc 2.0
"""
import logging

LOGGER = logging.getLogger(__name__)


class NoAccess(Exception):
    """ called when you should not see this """


class JsonRpcException(Exception):
    """ Throw xception that capture message """

    def __init__(self, code, message):
        super().__init__(self, message)
        self.js_code = code
        self.js_message = message


class JsonRpcMixin:
    """ Common functions used by both ws and http """

    @classmethod
    def get_params(cls, content):
        """ split params into args and kwargs """
        if isinstance(content["params"], dict):
            return [], content["params"]
        if isinstance(content["params"], list):
            return content["params"], {}
        raise JsonRpcException(-32602, "Params neither list or dict")

    async def handle_content(self, content):  # noqa: 901
        """ Assumes a json 2.0 method conent """
        LOGGER.info(content)
        result = {"jsonrpc": "2.0"}
        try:
            if content.get("jsonrpc") != "2.0":
                raise JsonRpcException(-32600, "protocol not supported")
            ref = content.get("id", None)
            if ref:
                result["id"] = ref

            method = content.get("method")
            if method is None:
                raise JsonRpcException(-32600, "no method")
            if method[0] == "_":
                raise JsonRpcException(-32600, "method private")
            if method not in self.procedures.__all__:
                raise JsonRpcException(-32600, f"no such method: {method}")

            user = self.current_user
            args, kwargs = self.get_params(content)
            LOGGER.debug("%s %s %s %s", user, method, args, kwargs)

            result["result"] = await self.application.perform(
                user, method, *args, **kwargs
            )
            if ref is None:
                result = None  # its a notification

        except JsonRpcException as ex:
            LOGGER.exception(ex)
            result["error"] = {"code": ex.js_code, "message": ex.js_message}
        except Exception as ex:  # pylint: disable=W0703
            LOGGER.exception(ex)
            result["error"] = {"code": -32000, "message": str(ex)}
        LOGGER.debug(repr(result))
        return result



"""
    This is a our invoke program
"""
import os
import sys
import tornado.log
from invoke import Program, Collection
from . import run
from . import db
from . import cfg

sys.path.insert(0, os.getcwd())

VERSION = "0.0.10"

_NAMESPACE_ = Collection()

_NAMESPACE_.add_task(run.run)
_NAMESPACE_.add_task(run.worker)
_NAMESPACE_.add_task(run.create)
_NAMESPACE_.add_task(db.upgrade)
_NAMESPACE_.add_task(db.downgrade)
_NAMESPACE_.add_task(db.revise)
_NAMESPACE_.add_task(cfg.cfg)

tornado.log.enable_pretty_logging()

program = Program(version=VERSION, namespace=_NAMESPACE_)  # pylint: disable=C0103



"""
    Worker listens to a queue in redis and performs work
"""
import asyncio
import logging
import uuid
from importlib import import_module
from concurrent.futures import ThreadPoolExecutor
import aioredis
import tornado.ioloop
import tornado.log
from . import config
from . import handlers
from .handlers import json_utils

LOGGER = logging.getLogger(__name__)


class RemoteException(Exception):
    """ We cannot serialise Exceptions so this is the message """


class Worker:
    """ A context rpc fed by redis """

    def __init__(self, cfg: config.Config, loop=None):
        """ setup variables before run """
        self.loop = loop if loop else tornado.ioloop.IOLoop.current()
        self.loop.set_default_executor(ThreadPoolExecutor(max_workers=cfg.max_workers))
        self.procedures = import_module(cfg.procedures)
        self.redis_url = cfg.redis_url
        self.redis_topic = cfg.redis_topic
        self.redis = None
        handlers.context.LOOP = self.loop
        handlers.BroadcastMixin._loop_ = self.loop
        handlers.BroadcastMixin._broadcaster_ = self
        self.loop.call_later(0, self.subscribe, cfg.redis_queue)

    async def shutdown(self):
        """ shut down outstatnding tasks """
        handlers.BroadcastMixin._broadcaster_ = None  # pylint: disable=W0212
        if self.redis:
            self.redis.close()
            await self.redis.wait_closed()

    async def subscribe(self, work_queue: str):
        """ run forever coroutine that listens to topic and broadcasts """
        subscription = await aioredis.create_redis(self.redis_url)
        try:
            while True:
                document = await subscription.blpop(work_queue, encoding="utf-8")
                LOGGER.info("got document: %s", document)
                self.loop.call_later(0, self.handle_work, document)
        except asyncio.CancelledError:
            pass

    async def unsubscribe(self):
        """ interface """

    async def send(self, data, user_ids):
        """ sends to redis topic """
        if self.redis is None:
            self.redis = await aioredis.create_redis_pool(self.redis_url)
        document = json_utils.dumps([data, user_ids])
        LOGGER.info("published broadcast %s -> %r", self.redis_topic, document)
        await self.redis.publish(self.redis_topic, document)

    async def handle_work(self, val: str = None):
        """ unpack the mesage and call perform """
        if self.redis is None:
            self.redis = await aioredis.create_redis_pool(self.redis_url)
        LOGGER.info("got work %s", val)
        if val is None:
            return
        content = json_utils.loads(val[1])
        try:
            proc = getattr(self.procedures, content["proc"])
            result = await handlers.context.perform(
                content["user"], proc, *content["args"], **content["kwargs"]
            )
            content["result"] = result
        except asyncio.CancelledError:
            LOGGER.info("cancelled")
            return
        except Exception as ex:  # pylint: disable=W0703
            content["error"] = {"code": -32000, "message": str(ex)}
        if content.get("reply"):
            document = json_utils.dumps(content)
            await self.redis.publish(content["reply"], document)
            LOGGER.info("published reply %s -> %r", content["reply"], document)


class Channel:
    """ sends work to redis and handles reply """

    def __init__(self, url, queue, loop=None):
        self.loop = loop if loop else tornado.ioloop.IOLoop.current()
        self._futures_ = {}
        self.redis_url = url
        self.redis_queue = queue
        self.reply_queue = f"channel:{uuid.uuid4()}"
        self.reply_task = self.loop.call_later(0, self.reply_subscribe)
        LOGGER.info("redis reply %s", self.reply_queue)
        self.redis = None

    async def tidy_up(self):
        """ a nasty little method to clean up an io_loop for testing """
        self.reply_task.cancel()
        if self.redis:
            self.redis.close()
            await self.redis.wait_closed()

    async def perform(self, user, proc, *args, **kwargs):
        """ send proc to redis """
        if self.redis is None:
            self.redis = await aioredis.create_redis_pool(self.redis_url)
        future_id = str(uuid.uuid4())
        future = asyncio.Future()
        self._futures_[future_id] = future
        document = handlers.json_utils.dumps(
            {
                "user": user,
                "proc": proc,
                "args": args,
                "kwargs": kwargs,
                "future_id": future_id,
                "reply": self.reply_queue,
            }
        )

        await self.redis.rpush(self.redis_queue, document)
        LOGGER.debug("published work %s -> %r", self.redis_queue, document)
        return await future

    async def reply_subscribe(self):
        """ run forever wait work from redis for Application """
        subscription = await aioredis.create_redis(self.redis_url)
        response = await subscription.subscribe(self.reply_queue)
        channel = response[0]
        LOGGER.info("subscribed to: %s", self.reply_queue)
        try:
            while await channel.wait_message():
                document = await channel.get(encoding="utf-8")
                LOGGER.info("got document: %s", document)
                result = handlers.json_utils.loads(document)
                future = self._futures_.get(result.get("future_id"), None)
                if future:
                    if "error" in result:
                        future.set_exception(RemoteException(result.get("error")))
                    else:
                        future.set_result(result.get("result"))
        except asyncio.CancelledError:
            pass


def main(cfg):
    """ Creates Worker and starts tornado """
    tornado.log.enable_pretty_logging()
    loop = tornado.ioloop.IOLoop.current()
    worker = Worker(cfg, loop)
    try:
        LOGGER.info("listening for work")
        loop.start()
    except KeyboardInterrupt:
        LOGGER.info("shutdown")
        loop.run_until_complete(worker.shutdown())
        loop.close()



# pylint: disable=W0201
"""
    Support for Tornado Authentication.
"""
import logging
from tornado.web import RequestHandler, HTTPError
from tornado.ioloop import IOLoop
from .user_mixin import UserMixin


class LogoutHandler(UserMixin, RequestHandler):  # pylint: disable=W0223
    """
        Removes the cookie from application settings
        and redirects.
    """

    def get(self):
        """ removes cookie and redirects to optional next argument """
        self.set_current_user(None)
        self.redirect(self.get_argument("next", "/"))


class LoginHandler(UserMixin, RequestHandler):  # pylint: disable=W0223
    """
        Can be called as ajax from the
        websocket client to get the auth cookie
        into the headers.
    """

    def initialize(self, login, register, page="login.html"):
        """ we're configured with functions to perform login and register """
        self.login = login
        self.register = register
        self.page = page

    def get(self, error=None, notice=None):
        """ render the form """
        email = self.get_argument("email", default=None)
        next_ = self.get_argument("next", "/")
        can_register = self.register is not None
        self.render(
            self.page,
            email=email,
            error=error,
            notice=notice,
            next=next_,
            can_register=can_register,
            app_name=self.settings.get("app_name", "liteblue"),
        )

    async def post(self):
        """ handle login post """
        try:
            email = self.get_argument("email", None)
            password = self.get_argument("password", None)
            submit = self.get_argument("submit", "login")
            if email is None or password is None:
                raise HTTPError(403, "email or password is None")

            loop = IOLoop.current()
            user = None
            if submit == "login":
                user = await loop.run_in_executor(None, self.login, email, password)
            elif self.register:
                user = await loop.run_in_executor(None, self.register, email, password)
            if user:
                self.set_current_user(user)
                self.redirect(self.get_argument("next", "/"))
            else:
                raise Exception("email or password incorrect")
        except Exception as ex:  # pylint: disable=W0703
            logging.exception(ex)
            self.get(error=str(ex))



"""
    tornado ws rpc server
"""
import logging
from importlib import import_module
from concurrent.futures import ThreadPoolExecutor
from tornado import web, ioloop
from . import handlers
from . import authentication
from .config import Config
from .worker import Channel

LOGGER = logging.getLogger(__name__)


class Application(web.Application):
    """ subclass of tornado Application """

    def __init__(self, cfg: Config, routes: list = None, io_loop: ioloop.IOLoop = None):
        """ called to initialize _broadcast_ attribute """
        procedures = import_module(cfg.procedures)
        settings = {
            k[8:]: getattr(cfg, k) for k in dir(cfg) if k.startswith("tornado_")
        }
        routes = (
            routes
            if routes
            else [
                (
                    r"/login",
                    handlers.LoginHandler,
                    {
                        "login": authentication.login,
                        "register": authentication.register,
                        "page": getattr(cfg, "login_page", "login.html"),
                    },
                ),
                (r"/logout", handlers.LogoutHandler),
                (r"/events", handlers.EventSource),
                (r"/rpc", handlers.RpcHandler),
                (r"/ws", handlers.RpcWebsocket),
                (
                    r"/(.*)",
                    handlers.AuthStaticFileHandler
                    if getattr(cfg, "tornado_login_url")
                    else web.StaticFileHandler,
                    {"default_filename": "index.html", "path": cfg.static_path},
                ),
            ]
        )
        super().__init__(routes, procedures=procedures, app_name=cfg.name, **settings)
        self._cfg_ = cfg
        self._loop_ = io_loop if io_loop else ioloop.IOLoop.current()
        self._loop_.set_default_executor(
            ThreadPoolExecutor(max_workers=cfg.max_workers)
        )
        handlers.context.LOOP = self._loop_
        handlers.BroadcastMixin.init_broadcasts(
            self._loop_, cfg.redis_topic, cfg.redis_url
        )
        if self._cfg_.redis_workers:
            self.channel = Channel(cfg.redis_url, cfg.redis_queue)

    async def tidy_up(self):
        """ a nasty little method to clean up an io_loop for testing """
        if self._cfg_.redis_workers:
            await self.channel.tidy_up()
        await handlers.BroadcastMixin.tidy_up()

    async def perform(self, user, proc, *args, **kwargs):
        """ runs a proc in threadpool or ioloop """
        if self._cfg_.redis_workers:
            return await self.channel.perform(user, proc, *args, **kwargs)
        proc = getattr(self.settings["procedures"], proc)
        return await handlers.context.perform(user, proc, *args, **kwargs)

    def run(self):  # pragma: no cover
        """ run the application """
        self.listen(self._cfg_.port)
        if self._cfg_.tornado_debug:
            LOGGER.info("running in debug mode")
        LOGGER.info("%s on port: %s", self._cfg_.name, self._cfg_.port)
        try:
            self._loop_.start()
        except KeyboardInterrupt:
            logging.info("shut down.")



""" tasks to create and run projects """
import os
import logging
from importlib import import_module
from pkg_resources import resource_filename
import tornado.template
from invoke import task
from .db import revise, upgrade
from .docker import docker
from .utils import qname_to_class, confirm_action, Colored

LOGGER = logging.getLogger(__name__)


@task(help={"project": "name of project"})
def create(ctx, project):
    """ creates a new project with a sqlite db """
    if os.path.isdir(project):
        if confirm_action(Colored.red("DELETE ") + project + "?"):
            ctx.run(f"rm -rf {project}")
        else:
            return
    if os.path.isfile(f"{project}.db"):
        os.unlink(f"{project}.db")
    LOGGER.info(Colored.cyan("creating: ") + project)
    template_dir = resource_filename("liteblue.apps", "simple")
    loader = tornado.template.Loader(template_dir)
    for path, folders, files in os.walk(template_dir):
        target_path = os.path.join(project, path[len(template_dir) + 1 :])
        for folder in folders:
            folder_path = os.path.join(target_path, folder)
            os.makedirs(folder_path)
        for filename in files:
            template = os.path.join(path, filename)
            target = os.path.join(target_path, os.path.splitext(filename)[0])
            tmpl = loader.load(template)
            with open(target, "wb") as file:
                file.write(tmpl.generate(project_name=project))
    os.makedirs(f"{project}/scripts/versions")
    revise(ctx, project, "first pass")
    upgrade(ctx, project, force=True)
    docker(ctx, project, force=True)


@task(help={"package": "the liteblue module to run"})
def run(_, package):
    """ run a liteblue project """
    app = import_module(package)
    app.main()


@task(help={"config": "config.Config module path"})
def worker(_, config):
    """ run a liteblue worker """
    from liteblue.worker import main  # pylint: disable=C0415

    cfg_ = qname_to_class(config)
    main(cfg_)



"""
    Simple mixin to provide utility methods
    to support User
"""
from .json_utils import dumps, loads


class UserMixin:
    """ Support for user in tornado handlers """

    @property
    def cookie_name(self):
        """ return the cookie_name declared in application settings"""
        return self.settings.get("cookie_name")

    def get_current_user(self):
        """ return the current user from the cookie """
        result = self.get_secure_cookie(self.cookie_name)
        if result:
            result = loads(result.decode("utf-8"))
        return result

    def set_current_user(self, value):
        """ put the current user in the cookie """
        if value:
            self.set_secure_cookie(self.cookie_name, dumps(value))
        else:
            self.clear_cookie(self.cookie_name)



# pylint: disable=C0103, W0401, W0614
""" sqlalchemy schema """
from sqlalchemy import *


metadata = MetaData()


user = Table(
    "user",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("email", String(128), nullable=False, unique=True, doc="required"),
    Column("password", String(64), doc="and no longer"),
)



""" tornado request handlers """
from .login_handler import LoginHandler, LogoutHandler
from .static_file_handler import AuthStaticFileHandler
from .rpc_websocket import RpcWebsocket
from .rpc_handler import RpcHandler
from .broadcast_mixin import BroadcastMixin
from .event_source import EventSource
